unit Pipes;

interface

uses
   SysUtils,Libc, IPCBase;

type
   TPipeBuffer = array of char;

   // Note if the order of these change, make sure changes
   // are reflected in the Close method.
   TPipeSide   = (psRead, psWrite);

   TPipe = class(TIPCBase)
     private
        // TODO: Check before releasing see if the function definition
        //       of pipe has changed to be more "user friendly"
        FHandles : array[0..1] of Integer;
     public
        constructor Create;
        destructor  Destroy; override;
        function Read(var buffer : TPipeBuffer; bufSize : Integer) : boolean;
        function Write(var buffer : TPipeBuffer; bufSize : Integer) : boolean;
        function Close(what : TPipeSide) : boolean;
        function WriteToStdOut : boolean;
        function ReadFromStdIn : boolean;
  end;

  TNamedPipe = class(TIPCBase)
    private
       FOpen     : boolean;
       FFileDesc : integer;
       FPipeName : string;
       FPipeSide : TPipeSide;

       function PipeSideToBits(pipeSide : TPipeSide) : integer;
    public
       constructor Create(name : string; pipeSide : TPipeSide;
                          perm : TIPCFilePermissions);
       destructor Destroy; override;
       function Read(var buffer : TPipeBuffer; bufSize : Integer) : boolean;
       function Write(var buffer : TPipeBuffer; bufSize : Integer) : boolean;
       function Close : boolean;
  end;

implementation

// Note: Possible future enhancment. Pipes can be opened in a Nonblock
//       mode as well. Use the fcntl API change the default blocking
//       mode to nonblocking


// Note: Another API that we could have looked at is popen.
//
// function popen(const Command : PChar; Modes : PChar) TIOFile; cdecl;
//
// Popen creates a pipe then creates another process (command).
// The parent then can either reads the standard output
// or writes to the standard input of the child process.

const
  IDX_READ_PIPE  =  0;
  IDX_WRITE_PIPE =  1;

constructor TNamedPipe.Create(name : string;
                              pipeSide : TPipeSide;
                              perm : TIPCFilePermissions);
var
  ret : integer;

begin
  inherited Create;

  FOpen := false;

  DebugMessage('calling mkfifo..');

  // create the pipe
  ret := mkfifo(PChar(name), FilePermissionsToBits(perm));
  // It's okay if the pipe already exists..
  if (ret < 0) and (GetLastError <> EEXIST) then
     ErrorMessage('mkfifo');

  FPipeName := name;
  FPipeSide := pipeSide;

  DebugMessage('calling open..');

  // now open the pipe
  FFileDesc := open(PChar(name), PipeSideToBits( FPipeSide ), 0);
  if FFileDesc <> LIBC_SUCCESS then
     ErrorMessage('NamedPipe.__open');

  DebugMessage('before return..');

  FOpen := true;
end;

destructor TNamedPipe.Destroy;
begin
  Close;
  inherited;
end;

function TNamedPipe.Read(var buffer : TPipeBuffer; bufSize : Integer) : boolean;
var
   ret : integer;

begin
   ret := __read(FFileDesc,buffer[0],bufSize);
   Result := (ret = LIBC_SUCCESS);
end;

function TNamedPipe.Write(var buffer : TPipeBuffer; bufSize : Integer) : boolean;
var
   ret : integer;

begin
   DebugMessage('writing data..');
   ret := __write(FFileDesc,buffer[0],bufSize);
   DebugMessage('flushing data..');
   // flush the write down the pipe..
   fsync(FFileDesc);
   Result := (ret = LIBC_SUCCESS);
end;

function TNamedPipe.Close : boolean;
begin
   if FOpen then
   begin
     __close(FFileDesc);
     FOpen := false;

     // we skip error checking because it will fail if another
     // process still has the file open. When the last process
     // ends, the pipe will be deleted.
     DeleteFile(FPipeName);
   end;
   Result := true;
end;

function TNamedPipe.PipeSideToBits(pipeSide : TPipeSide) : integer;
begin
  Result := 0;
  case pipeSide of
     psRead :  Result := O_RDONLY;
     psWrite : Result := O_WRONLY;
  else
     ErrorMessage('PipeSideToBits: unknown argument');
  end;
end;

constructor TPipe.Create;
var
  pptr : PInteger;
  ret : integer;

begin
  inherited;

  pptr   := @FHandles;
  ret    := pipe(pptr);
  if ret <> LIBC_SUCCESS then
     ErrorMessage('pipe');
  DebugMessage('The read size handle is ' + IntToStr(FHandles[IDX_READ_PIPE]) +
           ' the write side is ' + IntToStr(FHandles[IDX_WRITE_PIPE]));
end;

destructor TPipe.Destroy;
begin
  __close(FHandles[IDX_READ_PIPE]);
  __close(FHandles[IDX_WRITE_PIPE]);
  inherited;
end;

function TPipe.Read(var buffer : TPipeBuffer; bufSize : Integer) : boolean;
var
   ret : integer;

begin
   ret := __read(FHandles[IDX_READ_PIPE],buffer[0],bufSize);
   Result := (ret = LIBC_SUCCESS);
end;

function TPipe.Write(var buffer : TPipeBuffer; bufSize : Integer) : boolean;
var
   ret : integer;

begin
   ret := __write(FHandles[IDX_WRITE_PIPE],buffer[0],bufSize);
   // flush the write down the pipe..
   fsync(FHandles[IDX_WRITE_PIPE]);
   Result := (ret = LIBC_SUCCESS);
end;

function TPipe.Close(what : TPipeSide) : boolean;
begin
   writeln('Closing ',ord(what));
   __close(FHandles[ord(what)]);
   Result := true;
end;

function TPipe.WriteToStdOut : boolean;
var
   ret : integer;

begin
   DebugMessage('The read size handle is ' + IntToStr(FHandles[0]) +
            ' the write side is ' + IntToStr(FHandles[1]));

   ret := dup2(FHandles[IDX_WRITE_PIPE],STDOUT_FILENO);
   if (ret <> LIBC_SUCCESS) then
   begin
      DebugMessage('Uh oh dup2 failed ' + IntToStr(GetLastError));
      ErrorMessage('SendStdOutTotPipe:dup2');
   end;

   Result := (ret = LIBC_SUCCESS);
end;

function TPipe.ReadFromStdIn : boolean;
var
   ret : integer;

begin
   DebugMessage('The read size handle is ' + IntToStr(FHandles[0]) +
            ' the write side is ' + IntToStr(FHandles[1]));

   ret := dup2(FHandles[IDX_READ_PIPE],STDIN_FILENO);
   if (ret <> LIBC_SUCCESS) then
   begin
      DebugMessage('Uh oh send readStdInFromPipe failed (dup2)..' +
                IntToStr(GetLastError));
      ErrorMessage('ReadStdInFromPipe:dup2');
   end;

   Result := (ret = LIBC_SUCCESS);
end;

end.
